// Autor: Przemysaw Paprocki G1ISI

#include <iostream>
#include <vector>
#include <algorithm>
#include <functional>
#include <boost/lambda/lambda.hpp>
#include <boost/lambda/bind.hpp>
#include <boost/lambda/if.hpp>
#include <boost/lambda/switch.hpp>
#include <boost/lambda/loops.hpp>
#include <boost/lambda/casts.hpp>
#include <boost/lambda/exceptions.hpp>

/*Biblioteka standardowa algorithm udostpnia wiele przydatnych szablonw umoliwiajcych
wykonywanie czsto potrzebnych algorytmw na kontenerach. Konstrukcje te umoliwiaj np. wyszukiwanie,
sortowanie, przegladanie z modyfikacj itp. Problem w tym, e dla kadej operacji trzeba byo implementowa
drobn klas tzw.obiekt funkcyjny. Obiekt ten mia za zadanie dostarczy odpowiedniego operatora czy funkcji 
ktra miaa za zadanie okreli logik operacji na elementach kontenera. Nastpnie obiekt by przekawzywany 
do szablonu. Taka metoda powodowaa powstanie wiele obiektw funkcyjnych ktre niekiedy w duym projekcie
byy wykorzystywane tylko raz. Biblioteka boost wprowadza mechanizm (wraz z mechanizmami pomocniczymi) ktry 
umoliwia konstruowanie tak zwanych wyrae lambda. Mona je rozumie jako nienazwane funkcje ktre s wykonywane
w miejscu definicji. Ponisze przykady pokazuj (w niewielkim stopniu) moliwoci tych wyrae.*/

void funkcja_globalna (float i) 
{
	std::cout << "\nvoid funkcja_globalna : " << i;
}

struct example
{
	void funkcja_klasy (float i) const
	{
		std::cout << "\nvoid example::funkcja_klasy : " << i;
	}

	virtual ~example(){};
};

struct example_derived : public example 
{
	void funkcja_klasy_pochodnej (float i) const
	{
		std::cout << "void example_derived::funkcja_klasy_pochodnej : " << i << std::endl;
	}
};

int main()
{
	std::vector<float> vec_test;
 
    using namespace boost::lambda;

	// Przykad 1. Pierwszy rzut oka na wyraenie lambda :

	std::cout << "Przyklad 1:\n";
	(std::cout << _2 << " " << _1 << " " << _3 << "\n")				// Definicja wyraenia (to co sie wykona)
	(" zadna lala,","Zadna panna,", " nie zastapi terminala!\n");   // Argumenty wywoania.

	// W wyraeniach lambda argumenty oznaczamy jako _X gdzie X moe by cyfr od 1 do 9.
	// Bibloteka boost umoliwia zmian nazwy _X na dowoln poprzez uyci boost::lambda::placeholderX_type.
	// Jednak nie zaleca sie tego robi ze wzgldu na moliwo skonfudowania programistw uczestniczcych w
	// projekcie, a przyzwyczajonych do standardowego oznaczenia.

	// Przykad 2. Prezentacja starego dostpu do danych w algorytmach na kontenerze i odpowied boost.

	vec_test.push_back(0.5f);
	vec_test.push_back(3.4f);
	vec_test.push_back(666.666f);
	
	example ex;
	example * ex_ptr = &ex;

	std::cout << "Przyklad 2:";

	std::cout << "\nUzycie std::ptr_fun dla wiazania funkcji globalnej.";
	std::for_each(vec_test.begin(), vec_test.end(), std::ptr_fun(funkcja_globalna));

	std::cout << "\n\nWiazanie z metoda klasy za pomoca mem_fun_ref (dla obiektu ex).";
	std::for_each(vec_test.begin(), vec_test.end(), std::bind1st(std::mem_fun_ref(&example::funkcja_klasy), ex));

	std::cout << "\n\nWiazanie z metoda klasy za pomoca mem_fun (dla wskazania do obiektu ex).";
	std::for_each(vec_test.begin(), vec_test.end(), std::bind1st(std::mem_fun(&example::funkcja_klasy), ex_ptr));

	// Przy uywaniu metod z obecnego standardu musimy rozrni wizania dla obiektw i dla wskaza na te obiekty.
	// Wyraenia lambda z biblioteki boost "domylaj si" z czym maj doczynienia.

	std::cout << "\n\nUzycie boost::bind dla funkcji globalnych.";
	std::for_each(vec_test.begin(), vec_test.end(), bind(&funkcja_globalna,_1));

	std::cout << "\n\nUzycie boost::bind dla funkcji obiektow klasy.";
	std::for_each(vec_test.begin(), vec_test.end(), bind(&example::funkcja_klasy, ex ,_1));

	std::cout << "\n\nUzycie boost::bind dla funkcji wskazan do obiektow klasy.";
	std::for_each(vec_test.begin(), vec_test.end(), bind(&example::funkcja_klasy, ex_ptr ,_1));

	std::cout << "\n\nUzycie wyrazenia lambda dla obiektow (czy tez wskazan do nich) wprost z std::vector.\n";
	std::for_each(vec_test.begin(), vec_test.end(), std::cout << constant("Wyrazenie boost::lambda: ") <<_1 << constant("\n")  );

	// Przykad 3. Funkcje sterujce oraz ptle w wyraeniach lambda.

	// Przekazywane do szablonw standardowych algorytmw funkcje s zazwyczaj bardziej skomplikowane. 
	// Biblioteka boost::lambda umoliwia  standardowe konstrukcje sterujce wymienione poniej.

	// Przykad 3a. Konstrukcje typu if else

	std::cout << "\nPrzyklad 3a.\nUzycie konstrukcji if_else_then(warunek, instrukcje_dla_true, instrukcje_dla_false)." <<
		" Alternatywnie if_(warunek)[instrukcje_dla_true].else_[instrukcje_dla_false].\n";

	for_each(vec_test.begin(), vec_test.end(), 
		if_then_else(_1 > 3 , 
		std::cout << constant("\n(if_then_else) Liczba : ") << _1 << constant( " jest wieksza niz 3!\n") ,
		std::cout << constant("\n(if_then_else) Liczba : ") << _1 << constant( " jest mniejsza lub rowna 3!\n")));

  	for_each(vec_test.begin(), vec_test.end(), 
		if_(_1 > 3)
		[ 
			std::cout << constant("\n(if_().else_() Liczba : ") << _1 << constant( " jest wieksza niz 3!\n")
		].else_
		[
			std::cout << constant("\n(if_().else_() Liczba : ") << _1 << constant( " jest mniejsza lub rowna 3!\n")
		] );

	// Przykad 3b. Konstrukcje switch.

	 std::cout << "\nPrzyklad 3b.\nUzycie konstrukcji switch.\n\n"   ;
	 (switch_statement(
		 _1,
		 case_statement<0>
		 (var(std::cout) << constant("Wybrales 0 \n")),
		 case_statement<1>
		 (var(std::cout) << constant("Wybrales 1 \n")),
		 default_statement
		 (var(std::cout) << constant("Nie mam pojecia co wybrales!\n")))
	 ) ((make_const(666)));

	 // Uzycie (make_const(666)) jest konieczne. Genruje one const referencje przyjmonwan przez szablony boost.
	 // Ponadto cao jest wyraeniem lambda, dlatego aby byo dobrze zrozumiane prze bibliotek
	 // naley zmieni obiekt std::cout na element wyraenia lambda przy pomocy funkcji var().

	 // Przykad 3c. Przykad uycia ptli while.

	 std::cout << "\nPrzyklad 3c:\nUzycie konstrukcji (while_loop(warunek, instrukcje))(argumenty)\n\n";

	 int start = 5;
	 int end = 10;
	 (while_loop(_1 <= _2, 
		 (++_1, std::cout  << _2 << constant(" - ") << _1 << constant(" = ") << _2 -_1 << constant("\n") )))
		 (start,end);

	 // Przykad 3d. Przykad uycia ptli do_while

	 (do_while_loop( _1 == 0 , std::cout << constant ("\nPrzyklad 3d: \nUzycie konstrukcji (do_while_loop(warunek, instrukcje))(argumenty).\n")))(make_const(1));

	 (do_[
		 _1 == 0 , std::cout << constant ("\nUzycie konstrukcji (do_[instrukcje].while_(warunek))(argumenty).\n")
		 ].while_(_1 == 0))(make_const(1));

	 // Przykad 3e. Przykad uycia ptli for.

	 std::cout << "\nPrzyklad 3e.\nUzycie konstrukcji (for_ (war_poczatkowe,warunek, instrukcja iteracji)[instrukcje])(argumenty) : \n\n";
	 int i = 0;

	 (for_ (var(i)=0 , var(i) < _1, ++var(i))
		 [
			 var(std::cout) << var(i) << constant("^2 = ") << make_const( var(i)*var(i) ) << constant("\n")   
		 ]
	 )(make_const (4));

	 // Przykad 4. Biblioteka boost umoliwia wprowadzenie rzutowania w wyraeniu lambda jak rwnie
	 // genrowanie i przechwytywanie wyjtkw.

	 example_derived ex_derv;								    

	 // Rzutowanie w tym przykadzie nie powiedzie si i zostanie wygenerowany wyjtek, a nastpnie 
	 // przechwycony i obsuony.

	 std::cout << "\nPrzyklad4.\nUzycie rzutowania i wykorzystanie mechanizmu wyjatkow w wyrazeniach lambda.\n";
	 (try_catch(
		 bind(&example_derived::funkcja_klasy_pochodnej, ll_dynamic_cast<example_derived&>(*_1),_2),
		 catch_exception<std::bad_cast>(bind(&example::funkcja_klasy,_1,_2))))(ex_ptr, make_const(66.6f));	

	 std::cout << std::endl;
	 // Rzutowanie w tym przykadzie powiedzie si. Przedstawiono dodatkowo schemat zapisu przechwytywania
	 // wyjtkw rnych typw oraz dowolonego wyjtku.

	 (try_catch(
		 bind(&example_derived::funkcja_klasy_pochodnej, ll_dynamic_cast<example_derived&>(*_1),_2),
		 catch_exception<std::bad_cast>(bind(&example::funkcja_klasy,_1,_2)),
		 catch_exception<std::exception>(),
		 catch_all()))(make_const(&ex_derv), make_const(77.7f));

	 // Wyraenia lambda to przydatne konstrukcje ktre nie raz mog zaoszczdzi czas, zmniejszy ilo kodu czy
	 // zajt przez program pami. Mankamentem jest skadnia wyrae ktrej naley si po prostu nauczy i przewiczy
	 // na wielu przykadach.

	return 1;
}
